#pragma once

#include <vector>

#include <Ice/Handle.h>

#include <RobotAPI/libraries/armem/core/MemoryID.h>
#include <RobotAPI/interface/armem/query.h>


namespace armarx::armem::client::query::detail
{

    template <class _DerivedT, class _QueryT>
    class ChildSelectorOps
    {
    public:
        using DerivedT = _DerivedT;
        using QueryT = _QueryT;

        ChildSelectorOps() = delete;

        ChildSelectorOps(const armem::query::data::QueryTargets& p) :
            _parentTargets(p)
        {}


        virtual DerivedT& all() = 0;
        virtual DerivedT& withID(const MemoryID& id) = 0;


        void addQuery(const IceInternal::Handle<QueryT>& query)
        {
            _queries.push_back(query);
        }
        void addQueries(const std::vector<IceInternal::Handle<QueryT>>& queries)
        {
            for (auto q : queries)
            {
                addQuery(q);
            }
        }

    protected:
        armem::query::data::QueryTargets _parentTargets;


    protected:

        template <class DerivedQueryT>
        DerivedQueryT& _addQuery()
        {
            IceInternal::Handle<DerivedQueryT> query(new DerivedQueryT);
            _queries.push_back(query);
            return *query;
        }

        std::vector<QueryT> _queries;


        // Only friends have access to these.

        template <class ParentDerivedT, class ParentChildT>
        friend class ParentSelectorOps;

        template <class T>
        inline void _apply(T arg)
        {
            arg(dynamic_cast<DerivedT&>(*this));
        }
        template <class T, class ...Ts>
        inline void _apply(T arg, Ts... args)
        {
            _apply(arg);
            _apply(args...);
        }



    };



    template <class _DerivedT, class _ChildT>
    class ParentSelectorOps
    {
    public:
        using DerivedT = _DerivedT;
        using ChildT = _ChildT;


    public:
        virtual ~ParentSelectorOps() = default;

        /// QueryTargets for this query
        DerivedT& queryTargets(const armem::query::data::QueryTargets& targets)
        {
            this->_targets = targets;
            return dynamic_cast<DerivedT&>(*this);
        }

    protected:

        ChildT& _addChild()
        {
            return _children.emplace_back(_targets);
        }
        ChildT& _addChild(const ChildT& child)
        {
            return _children.emplace_back(child);
        }

        template <class ...Ts>
        ChildT& _addChild(Ts... args)
        {
            ChildT& sel = _addChild();
            sel._apply(args...);
            return sel;
        }


        std::vector<ChildT> _children;
        armem::query::data::QueryTargets _targets = {armem::query::data::QueryTarget::WM}; // if not specified we only query the WM
    };



    template <class DerivedT, class QueryT, class ChildT>
    class InnerSelectorOps :
        public ParentSelectorOps<DerivedT, ChildT>,
        public ChildSelectorOps<DerivedT, QueryT>
    {
    public:

        InnerSelectorOps(const armem::query::data::QueryTargets& p) :
            ChildSelectorOps<DerivedT, QueryT>(p)
        {}

        virtual std::vector<QueryT> buildQueries() const
        {
            // check own query targets and parent query targets
            for (const auto& target : this->_targets)
            {
                if (std::find(this->_parentTargets.begin(), this->_parentTargets.end(), target) == this->_parentTargets.end())
                {
                    // TODO: Error, this makes no sense! or just ignore?
                }
            }

            std::vector<typename ChildT::QueryT> childQueries;
            for (const auto& child : this->_children)
            {
                for (auto& query : child.buildQueries())
                {
                    query->targets = this->_targets;
                    childQueries.push_back(query);
                }
            }

            std::vector<QueryT> queries = this->_queries;
            for (auto& query : queries)
            {
                this->_setChildQueries(query, childQueries);
            }
            return queries;
        }


    protected:

        virtual void _setChildQueries(QueryT& query, const std::vector<typename ChildT::QueryT>& childQueries) const = 0;

    };

}
