/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    RobotAPI::ArmarXObjects::armem
 * @author     Simon Ottenhaus ( simon dot ottenhaus at kit dot edu )
 * @date       2020
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#define BOOST_TEST_MODULE RobotAPI::ArmarXLibraries::armem

#define ARMARX_BOOST_TEST

#include <RobotAPI/Test.h>
#include <RobotAPI/libraries/armem/core/Commit.h>
#include <RobotAPI/libraries/armem/core/wm/memory_definitions.h>
#include <RobotAPI/libraries/armem/core/operations.h>

#include <iostream>
#include <set>


namespace armem = armarx::armem;


template <class ValueT>
static
std::vector<ValueT> toVector(const std::set<ValueT>& set)
{
    return {set.begin(), set.end()};
}



BOOST_AUTO_TEST_CASE(test_forEach)
{
    using namespace armarx::armem;

    const MemoryID mid("memory");
    const std::set<MemoryID> emptySet;  // For comparison.

    wm::Memory memory(mid);
    std::set<MemoryID> cids, pids, eids, sids, iids;

    for (size_t c = 0; c <= 2; ++c)
    {
        const MemoryID cid = mid.withCoreSegmentName("core_" + std::to_string(c));
        cids.insert(cid);
        wm::CoreSegment& coreSeg = memory.addCoreSegment(cid.coreSegmentName);
        for (size_t p = 0; p <= c; ++p)
        {
            const MemoryID pid = cid.withProviderSegmentName("prov_" + std::to_string(p));
            pids.insert(pid);
            for (size_t e = 0; e <= p; ++e)
            {
                const MemoryID eid = pid.withEntityName("entity_" + std::to_string(e));
                eids.insert(eid);
                for (size_t s = 0; s <= e; ++s)
                {
                    const MemoryID sid = eid.withTimestamp(Time::microSeconds(int(s)));
                    sids.insert(sid);

                    EntityUpdate update;
                    update.entityID = eid;
                    update.timeCreated = sid.timestamp;
                    for (size_t i = 0; i <= s; ++i)
                    {
                        const MemoryID iid = sid.withInstanceIndex(int(i));
                        iids.insert(iid);
                        update.instancesData.emplace_back();
                    }
                    auto r = memory.update(update);

                    BOOST_CHECK(coreSeg.hasProviderSegment(pid.providerSegmentName));
                }
            }

            BOOST_CHECK_EQUAL(coreSeg.size(), p + 1);
        }
        BOOST_CHECK_EQUAL(memory.size(), c + 1);
    }

    BOOST_CHECK_GT(iids.size(), 0);
    BOOST_CHECK_GT(sids.size(), 0);
    BOOST_CHECK_GT(eids.size(), 0);
    BOOST_CHECK_GT(pids.size(), 0);
    BOOST_CHECK_GT(cids.size(), 0);

    BOOST_TEST_MESSAGE("Memory: \n" << armem::print(memory));

    memory.forEachInstance([&](const wm::EntityInstance & i) -> bool
    {
        BOOST_TEST_CONTEXT(wm::EntityInstance::getLevelName() << " " << i.id())
        {
            BOOST_TEST_INFO(toVector(iids));
            BOOST_CHECK_EQUAL(iids.count(i.id()), 1);
            iids.erase(i.id());
        }
        return true;
    });
    BOOST_TEST_CONTEXT(toVector(iids))
    {
        BOOST_CHECK_EQUAL(iids.size(), 0);
    }

    memory.forEachSnapshot([&](const wm::EntitySnapshot & s) -> bool
    {
        BOOST_TEST_CONTEXT(wm::EntitySnapshot::getLevelName() << " " << s.id())
        {
            BOOST_TEST_INFO(toVector(sids));
            BOOST_CHECK_EQUAL(sids.count(s.id()), 1);
            sids.erase(s.id());
        }
        return true;
    });
    BOOST_TEST_INFO(toVector(sids));
    BOOST_CHECK_EQUAL(sids.size(), 0);

    memory.forEachEntity([&](const wm::Entity & e) -> bool
    {
        BOOST_TEST_CONTEXT(wm::Entity::getLevelName() << " " << e.id())
        {
            BOOST_TEST_INFO(toVector(eids));
            BOOST_CHECK_EQUAL(eids.count(e.id()), 1);
            eids.erase(e.id());
        }
        return true;
    });
    BOOST_TEST_INFO(toVector(eids));
    BOOST_CHECK_EQUAL(eids.size(), 0);

    memory.forEachProviderSegment([&](const wm::ProviderSegment & p)  // -> bool
    {
        BOOST_TEST_CONTEXT(wm::ProviderSegment::getLevelName() << " " << p.id())
        {
            BOOST_TEST_INFO(toVector(pids));
            BOOST_CHECK_EQUAL(pids.count(p.id()), 1);
            pids.erase(p.id());
        }
        return true;
    });
    BOOST_TEST_INFO(toVector(pids));
    BOOST_CHECK_EQUAL(pids.size(), 0);

    memory.forEachCoreSegment([&](const wm::CoreSegment & c)  // -> bool
    {
        BOOST_TEST_CONTEXT(wm::CoreSegment::getLevelName() << " " << c.id())
        {
            BOOST_TEST_INFO(toVector(cids));
            BOOST_CHECK_EQUAL(cids.count(c.id()), 1);
            cids.erase(c.id());
        }
        return true;
    });
    BOOST_TEST_INFO(toVector(cids));
    BOOST_CHECK_EQUAL(cids.size(), 0);

}


BOOST_AUTO_TEST_CASE(test_forEach_non_bool_func)
{
    // Check whether this compiles + runs without break.

    armem::wm::Entity entity;
    entity.addSnapshot(armem::Time::microSeconds(500));
    entity.addSnapshot(armem::Time::microSeconds(1500));

    int i = 0;
    entity.forEachSnapshot([&i](const armem::wm::EntitySnapshot&) -> void
    {
        ++i;
    });
    BOOST_CHECK_EQUAL(i, 2);
}
