#include "MemoryViewer.h"

#include <RobotAPI/libraries/armem/core/wm/ice_conversions.h>
#include <RobotAPI/libraries/armem_gui/gui_utils.h>

#include <RobotAPI/libraries/armem/server/query_proc/wm.h>
#include <RobotAPI/libraries/armem/server/query_proc/ltm.h>

#include <ArmarXGui/libraries/SimpleConfigDialog/SimpleConfigDialog.h>

#include <ArmarXCore/core/ManagedIceObject.h>
#include <ArmarXCore/core/time/TimeUtil.h>
#include <ArmarXCore/observers/variant/Variant.h>

#include <SimoxUtility/algorithm/get_map_keys_values.h>

#include <QApplication>
#include <QBoxLayout>
#include <QCheckBox>
#include <QClipboard>
#include <QDialog>
#include <QGroupBox>
#include <QMenu>
#include <QLabel>
#include <QLayout>
#include <QSettings>
#include <QTimer>

#include <iomanip>


namespace armarx::armem::gui
{
    MemoryViewer::MemoryViewer(
        QBoxLayout* updateWidgetLayout,
        QGroupBox* memoryGroupBox, QLayout* memoryGroupBoxParentLayout,
        QGroupBox* instanceGroupBox, QLayout* instanceGroupBoxParentLayout,
        QBoxLayout* diskControlWidgetLayout,
        QLabel* statusLabel
    )
    {
        Logging::setTag("MemoryViewer");

        this->statusLabel = statusLabel;
        this->statusLabel->clear();

        statusLabel->setContextMenuPolicy(Qt::CustomContextMenu);
        connect(statusLabel, &QLabel::customContextMenuRequested, [statusLabel](const QPoint & pos)
        {
            QMenu menu(statusLabel);
            menu.addAction("Copy to clipboard", [statusLabel]()
            {
                QApplication::clipboard()->setText(statusLabel->text());
            });
            menu.addAction("Clear status", [statusLabel]()
            {
                statusLabel->clear();
            });
            menu.exec(statusLabel->mapToGlobal(pos));
        });


        // Update timer
        this->updateWidgetLayout = updateWidgetLayout;
        updateWidget = new armem::gui::PeriodicUpdateWidget(2.0, 60);
        updateWidgetLayout->insertWidget(0, updateWidget);

        processQueryResultTimer = new QTimer(this);
        processQueryResultTimer->setInterval(1000 / 60);  // Keep this stable.
        processQueryResultTimer->start();

        // Memory View
        memoryGroup = new armem::gui::MemoryGroupBox();
        armarx::gui::replaceWidget(memoryGroupBox, memoryGroup, memoryGroupBoxParentLayout);
        ARMARX_CHECK_NULL(memoryGroupBox);

        // Instance View
        this->instanceGroup = new armem::gui::InstanceGroupBox();
        armarx::gui::replaceWidget(instanceGroupBox, instanceGroup, instanceGroupBoxParentLayout);
        this->instanceGroup->setStatusLabel(statusLabel);
        ARMARX_CHECK_NULL(instanceGroupBox);

        // Disk Control
        if (diskControlWidgetLayout)
        {
            this->diskControlLayout = diskControlWidgetLayout;
            diskControl = new armem::gui::disk::ControlWidget();
            diskControlWidgetLayout->addWidget(diskControl);
        }

        // Connections
        //connect(this, &This::connected, this, &This::updateMemory);

        connect(diskControl, &armem::gui::disk::ControlWidget::requestedStoreOnDisk, this, &This::storeOnDisk);
        connect(diskControl, &armem::gui::disk::ControlWidget::requestedLoadFromDisk, this, &This::loadFromDisk);

        connect(this, &This::connected, this, &This::startQueries);
        connect(updateWidget, &armem::gui::PeriodicUpdateWidget::update, this, &This::startQueries);
        connect(processQueryResultTimer, &QTimer::timeout, this, &This::processQueryResults);

        connect(memoryGroup->queryWidget(), &armem::gui::QueryWidget::storeInLTM, this, &This::storeInLTM);

        connect(this, &This::memoryDataChanged, this, &This::updateMemoryTree);
        connect(memoryGroup->tree(), &armem::gui::MemoryTreeWidget::selectedItemChanged, this, &This::updateInstanceTree);

        connect(memoryGroup->tree(), &armem::gui::MemoryTreeWidget::updated, this, &This::memoryTreeUpdated);
        connect(instanceGroup, &armem::gui::InstanceGroupBox::viewUpdated, this, &This::instanceTreeUpdated);
        connect(instanceGroup->view, &armem::gui::InstanceView::memoryIdResolutionRequested, this, &This::resolveMemoryID);
    }


    void MemoryViewer::setLogTag(const std::string& _tag)  // Leading _ silences a warning
    {
        Logging::setTag(_tag);
    }


    void MemoryViewer::onInit(ManagedIceObject& component)
    {
        if (mnsName.size() > 0)
        {
            component.usingProxy(mnsName);
        }
        if (debugObserverName.size() > 0)
        {
            component.usingProxy(debugObserverName);
        }

        emit initialized();
    }


    void MemoryViewer::onConnect(ManagedIceObject& component)
    {
        if (not mnsName.empty())
        {
            armem::mns::MemoryNameSystemInterfacePrx mnsProxy;
            component.getProxy(mnsProxy, mnsName);
            mns = client::MemoryNameSystem(mnsProxy);

            const bool update = true;
            memoryReaders = mns.getAllReaders(update);
        }
        // DebugObserver is optional (check for null on every call)
        if (not debugObserverName.empty())
        {
            component.getProxy(debugObserver, debugObserverName, false, "", false);
        }
        updateWidget->startTimerIfEnabled();

        emit connected();
    }


    void MemoryViewer::onDisconnect(ManagedIceObject&)
    {
        updateWidget->stopTimer();

        emit disconnected();
    }


    const armem::wm::Memory* MemoryViewer::getSingleMemoryData(const std::string& memoryName)
    {
        auto it = memoryData.find(memoryName);
        if (it == memoryData.end())
        {
            std::stringstream ss;
            ss << "Memory name '" << memoryName << "' is unknown. Known are: "
               << simox::alg::get_keys(memoryData);
            statusLabel->setText(QString::fromStdString(ss.str()));
            return nullptr;
        }
        /*else if (not it->second.has_value())
        {
            return nullptr;
        }*/
        else
        {
            //return &it->second.value();
            return &it->second;
        }
    }


    void MemoryViewer::storeInLTM()
    {
        TIMING_START(MemoryStore)

        for (auto& [name, reader] : memoryReaders)
        {
            data::StoreInput input;
            input.query = memoryGroup->queryWidget()->queryInput().toIce();
            reader.readAndStore(input);
        }

        TIMING_END_STREAM(MemoryStore, ARMARX_VERBOSE)
    }


    void MemoryViewer::storeOnDisk(QString directory)
    {
        TIMING_START(MemoryExport)

        std::string status;
        diskControl->storeOnDisk(directory, memoryData, &status);

        statusLabel->setText(QString::fromStdString(status));
        TIMING_END_STREAM(MemoryExport, ARMARX_VERBOSE)
    }


    void MemoryViewer::loadFromDisk(QString directory)
    {
        std::string status;
        std::map<std::string, wm::Memory> data =
            diskControl->loadFromDisk(directory, memoryGroup->queryWidget()->queryInput(), &status);

        for (auto& [name, memory] : data)
        {
            this->memoryData[name] = std::move(memory);
        }
        statusLabel->setText(QString::fromStdString(status));

        emit memoryDataChanged();
    }


    void MemoryViewer::startQueries()
    {
        memoryReaders = mns.getAllReaders(true);
        startDueQueries(runningQueries, memoryReaders);
    }


    void MemoryViewer::processQueryResults()
    {
        const std::map<std::string, client::QueryResult> results = collectQueryResults(runningQueries, memoryReaders);

        int errorCount = 0;
        applyQueryResults(results, &errorCount);

        emit memoryDataChanged();
        updateStatusLabel(errorCount);
    }


    void MemoryViewer::updateStatusLabel(int errorCount)
    {
        // Code to output status label information
        if (statusLabel and errorCount > 0)
        {
            auto now = std::chrono::system_clock::now();
            auto in_time_t = std::chrono::system_clock::to_time_t(now);

            std::stringstream ss;
            ss << "Last update: " << std::put_time(std::localtime(&in_time_t), "%Y-%m-%d %X");
            ss << "\nThe query produced " << errorCount << " errors! Please check log.";

            statusLabel->setText(QString::fromStdString(ss.str()));
        }
    }


    void MemoryViewer::startDueQueries(
        std::map<std::string, Ice::AsyncResultPtr>& queries,
        const std::map<std::string, armem::client::Reader>& readers)
    {
        armem::client::QueryInput input = memoryGroup->queryWidget()->queryInput();

        for (auto& [name, reader] : readers)
        {
            if (queries.count(name) == 0)
            {
                queries[name] = reader.memoryPrx->begin_query(input.toIce());
            }
        }
    }


    std::map<std::string, client::QueryResult>
    MemoryViewer::collectQueryResults(
        std::map<std::string, Ice::AsyncResultPtr>& queries,
        const std::map<std::string, client::Reader>& readers)
    {
        TIMING_START(tCollectQueryResults)

        std::map<std::string, client::QueryResult> results;
        for (auto it = queries.begin(); it != queries.end();)
        {
            const std::string& name = it->first;
            Ice::AsyncResultPtr& queryPromise = it->second;

            if (queryPromise->isCompleted())
            {
                if (auto jt = memoryReaders.find(name); jt != readers.end())
                {
                    try
                    {
                        results[name] = client::QueryResult::fromIce(jt->second.memoryPrx->end_query(queryPromise));
                    }
                    catch (const Ice::ConnectionRefusedException&)
                    {
                        // Server is gone (MNS did not now about it yet) => Skip result.
                    }
                }
                // else: Server is gone (MNS new about it) => Skip result.

                // Promise is completed => Clean up in any case.
                it = runningQueries.erase(it);
            }
            else
            {
                ++it;  // Uncompleted => Keep.
            }
        }

        TIMING_END_STREAM(tCollectQueryResults, ARMARX_VERBOSE)
        if (debugObserver)
        {
            debugObserver->begin_setDebugChannel(Logging::tag.tagName,
            {
                { "t Collect Query Results [ms]", new Variant(tCollectQueryResults.toMilliSecondsDouble()) },
                { "# Collected Query Results", new Variant(static_cast<int>(results.size())) },
            });
        }

        return results;
    }


    void MemoryViewer::applyQueryResults(
        const std::map<std::string, client::QueryResult>& results,
        int* outErrorCount)
    {
        TIMING_START(tProcessQueryResults)
        for (const auto& [name, result] : results)
        {
            if (result.success)
            {
                memoryData[name] = std::move(result.memory);
            }
            else
            {
                ARMARX_WARNING << "Querying memory server '" << name << "' produced an error: \n" << result.errorMessage;
                if (outErrorCount)
                {
                    outErrorCount++;
                }
            }
        }

        // Drop all entries in memoryData which are not in memoryReaders anymore.
        for (auto it = memoryData.begin(); it != memoryData.end();)
        {
            if (memoryReaders.count(it->first) == 0)
            {
                it = memoryData.erase(it);
            }
            else
            {
                ++it;
            }
        }

        TIMING_END_STREAM(tProcessQueryResults, ARMARX_VERBOSE)
        if (debugObserver)
        {
            debugObserver->begin_setDebugChannel(Logging::tag.tagName,
            {
                { "t Process Query Results [ms]", new Variant(tProcessQueryResults.toMilliSecondsDouble()) },
                { "# Processed Query Results", new Variant(static_cast<int>(results.size())) },
            });
        }
    }


    void MemoryViewer::updateInstanceTree(const armem::MemoryID& selectedID)
    {
        const armem::wm::Memory* data = getSingleMemoryData(selectedID.memoryName);
        if (data)
        {
            if (not selectedID.hasEntityName())
            {
                return;
            }
            armem::MemoryID id = selectedID;
            const armem::wm::EntitySnapshot* snapshot = nullptr;
            if (not id.hasTimestamp())
            {
                const armem::wm::Entity& entity = data->getEntity(id);
                if (entity.empty())
                {
                    return;
                }
                snapshot = &entity.getLatestSnapshot();
                id.timestamp = snapshot->time();
            }
            if (not id.hasInstanceIndex())
            {
                if (not snapshot)
                {
                    try
                    {
                        snapshot = &data->getSnapshot(id);
                    }
                    catch (const armem::error::ArMemError& e)
                    {
                        if (statusLabel)
                        {
                            statusLabel->setText(e.what());
                        }
                    }
                }
                if (snapshot && snapshot->size() > 0)
                {
                    id.instanceIndex = 0;
                }
            }
            if (id.hasInstanceIndex())
            {
                instanceGroup->view->update(id, *data);
            }
        }
    }


    void MemoryViewer::resolveMemoryID(const MemoryID& id)
    {
        // ARMARX_IMPORTANT << "Resolving memory ID: " << id;

        auto handleError = [this](const std::string & msg)
        {
            statusLabel->setText(QString::fromStdString(msg));
            ARMARX_WARNING << msg;
        };

        if (id.memoryName.empty())
        {
            handleError("Memory name is empty.");
        }

        aron::typenavigator::ObjectNavigatorPtr segmentType;
        std::optional<wm::EntityInstance> instance;
        try
        {
            if (const wm::Memory* data = getSingleMemoryData(id.memoryName))
            {
                segmentType = data->getProviderSegment(id).aronType();

                if (id.hasInstanceIndex())
                {
                    instance = data->getInstance(id);
                }
                else if (id.hasTimestamp())
                {
                    instance = data->getSnapshot(id).getInstance(0);
                }
                else
                {
                    instance = data->getEntity(id).getLatestSnapshot().getInstance(0);
                }
            }
        }
        catch (const armem::error::ArMemError&)
        {
            // May be handled by remote lookup
        }

        if (not instance)
        {
            try
            {
                // Resolve remotely (may still fail, returns an optional).
                instance = mns.resolveEntityInstance(id);
            }
            catch (const armem::error::ArMemError& e)
            {
                ARMARX_WARNING << e.what();
                statusLabel->setText(e.what());
            }
        }

        if (instance)
        {
            instanceGroup->view->addInstanceView(*instance, segmentType);
        }
        else
        {
            // ToDo: Propagate error back to highlight selected entry in red
        }
    }


    void MemoryViewer::updateMemoryTree()
    {
        std::map<std::string, const armem::wm::Memory*> convMap;
        for (auto& [name, data] : memoryData)
        {
            convMap[name] = &data;
        }

        // if convMap.empty(), we still need to update to remove existing entries.

        TIMING_START(GuiUpdate)
        memoryGroup->tree()->update(convMap);
        TIMING_END_STREAM(GuiUpdate, ARMARX_VERBOSE)

        if (debugObserver)
        {
            try
            {
                debugObserver->setDebugDatafield(Logging::tag.tagName, "GUI Update [ms]", new Variant(GuiUpdate.toMilliSecondsDouble()));
            }
            catch (const Ice::Exception&)
            {
                // Ignore ...
            }
        }
    }


    const static std::string CONFIG_KEY_MEMORY = "MemoryViewer.MemoryNameSystem";
    const static std::string CONFIG_KEY_DEBUG_OBSERVER = "MemoryViewer.DebugObserverName";


    void MemoryViewer::loadSettings(QSettings* settings)
    {
        mnsName = settings->value(QString::fromStdString(CONFIG_KEY_MEMORY), "MemoryNameSystem").toString().toStdString();
        debugObserverName = settings->value(QString::fromStdString(CONFIG_KEY_DEBUG_OBSERVER), "DebugObserver").toString().toStdString();
    }
    void MemoryViewer::saveSettings(QSettings* settings)
    {
        settings->setValue(QString::fromStdString(CONFIG_KEY_MEMORY), QString::fromStdString(mnsName));
        settings->setValue(QString::fromStdString(CONFIG_KEY_DEBUG_OBSERVER), QString::fromStdString(debugObserverName));
    }


    void MemoryViewer::writeConfigDialog(SimpleConfigDialog* dialog)
    {
        dialog->addProxyFinder<armarx::armem::mns::MemoryNameSystemInterfacePrx>({CONFIG_KEY_MEMORY, "MemoryNameSystem", "MemoryNameSystem"});
        dialog->addProxyFinder<armarx::DebugObserverInterfacePrx>({CONFIG_KEY_DEBUG_OBSERVER, "Debug Observer", "DebugObserver"});
    }
    void MemoryViewer::readConfigDialog(SimpleConfigDialog* dialog)
    {
        mnsName = dialog->getProxyName(CONFIG_KEY_MEMORY);
        if (mnsName.empty())
        {
            mnsName = "MemoryNameSystem";
        }
        debugObserverName = dialog->getProxyName(CONFIG_KEY_DEBUG_OBSERVER);
    }

}
