/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2012-2016, High Performance Humanoid Technologies (H2T),
 * Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Fabian Peller-Konrad (fabian dot peller-konrad at kit dot edu)
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

// Header
#include "IntEnumClass.h"

#include <SimoxUtility/meta/type_name.h>


namespace armarx::aron::cppserializer::serializer
{
    // constructors
    IntEnumClassSerializer::IntEnumClassSerializer(const typenavigator::IntEnumNavigatorPtr& n) :
        detail::SerializerBase<typenavigator::IntEnumNavigator, IntEnumClassSerializer>(
            n->getEnumName(),
            simox::meta::get_type_name<data::AronNDArray>(),
            simox::meta::get_type_name<type::AronIntEnum>(),
            n)
    {
        ARMARX_CHECK_NOT_NULL(typenavigator);
        if (typenavigator->getMaybe() != type::Maybe::eNone)
        {
            throw error::MaybeNotValidException("IntEnumClassSerializer", "IntEnumClassSerializer", "Somehow the maybe flag of a top level int enum declaration is set. This is not valid!", typenavigator->getMaybe(), typenavigator->getPath());
        }
    }

    std::vector<CppFieldPtr> IntEnumClassSerializer::getPublicVariableDeclarations(const std::string&) const
    {
        std::vector<CppFieldPtr> fields;
        std::stringstream enum_to_name;
        std::stringstream name_to_enum;
        std::stringstream enum_to_value;
        std::stringstream value_to_enum;

        enum_to_name << "{" << std::endl;
        name_to_enum << "{" << std::endl;
        enum_to_value << "{" << std::endl;
        value_to_enum << "{" << std::endl;
        for (const auto& [key, value] : typenavigator->getAcceptedValueMap())
        {
            std::string enumKeyWithNamespace = std::string(IMPL_ENUM) + "::" + key;
            fields.push_back(std::make_shared<CppField>("const static " + std::string(IMPL_ENUM), key + " = " + enumKeyWithNamespace));

            enum_to_name << "\t\t{" << enumKeyWithNamespace << ", \"" << key << "\"}," << std::endl;
            name_to_enum << "\t\t{\"" << key << "\", " << enumKeyWithNamespace << "}," << std::endl;

            enum_to_value << "\t\t{" << enumKeyWithNamespace << ", " << value << "}," << std::endl;
            value_to_enum << "\t\t{" << value << ", " << enumKeyWithNamespace << "}," << std::endl;
        }

        enum_to_name << "\t}";
        name_to_enum << "\t}";
        enum_to_value << "\t}";
        value_to_enum << "\t}";

        fields.push_back(std::make_shared<CppField>("const std::map<" + std::string(IMPL_ENUM) + ", std::string>", "EnumToStringMap", enum_to_name.str(), "Mapping enum values to readable strings"));
        fields.push_back(std::make_shared<CppField>("const std::map<std::string, " + std::string(IMPL_ENUM) + ">", "StringToEnumMap", name_to_enum.str(), "Mapping readable strings to enum values"));
        fields.push_back(std::make_shared<CppField>("const std::map<" + std::string(IMPL_ENUM) + ", int>", "EnumToValueMap", enum_to_value.str(), "Mapping enum values to a int value"));
        fields.push_back(std::make_shared<CppField>("const std::map<int, " + std::string(IMPL_ENUM) + ">", "ValueToEnumMap", value_to_enum.str(), "Mapping int values to a enum"));

        fields.push_back(std::make_shared<CppField>(std::string(IMPL_ENUM), "value", "", "The current value of the enum object"));

        return fields;
    }

    CppBlockPtr IntEnumClassSerializer::getResetSoftBlock(const std::string& accessor) const
    {
        CppBlockPtr block_if_data = CppBlockPtr(new CppBlock());
        block_if_data->addLine("value = {};");
        return ResolveMaybeResetSoftBlock(accessor, block_if_data, typenavigator);
    }

    CppBlockPtr IntEnumClassSerializer::getResetHardBlock(const std::string& accessor) const
    {
        CppBlockPtr block_if_data = CppBlockPtr(new CppBlock());
        block_if_data->addLine("value = {};");
        return ResolveMaybeResetSoftBlock(accessor, block_if_data, typenavigator);
    }

    CppBlockPtr IntEnumClassSerializer::getWriteTypeBlock(const std::string&) const
    {
        CppBlockPtr b = CppBlockPtr(new CppBlock());
        b->addLine("w.writeInt({__aronMaybeType}); // of top level enum " + getCoreCppTypename());
        return b;
    }

    CppBlockPtr IntEnumClassSerializer::getWriteBlock(const std::string& accessor) const
    {
        CppBlockPtr block_if_data = CppBlockPtr(new CppBlock());
        block_if_data->addLine("w.writePrimitive(EnumToValueMap.at(value)); // of top level enum " + getCoreCppTypename());
        return ResolveMaybeWriteBlock(accessor, block_if_data, typenavigator);
    }

    CppBlockPtr IntEnumClassSerializer::getReadBlock(const std::string& accessor) const
    {
        CppBlockPtr block_if_data = CppBlockPtr(new CppBlock());
        block_if_data->addLine("int temporary;");
        block_if_data->addLine("r.readEndPrimitive(temporary); // of top level enum " + getCoreCppTypename());
        block_if_data->addLine("value = ValueToEnumMap.at(temporary);");
        return ResolveMaybeReadBlock(accessor, "r.readStartPrimitive()", block_if_data, typenavigator);
    }

    CppBlockPtr IntEnumClassSerializer::getEqualsBlock(const std::string& accessor, const std::string& otherInstanceAccessor) const
    {
        CppBlockPtr block_if_data = CppBlockPtr(new CppBlock());
        block_if_data->addLine("if (not (value == " + otherInstanceAccessor + ".value))");
        block_if_data->addLineAsBlock("return false;");
        return ResolveMaybeEqualsBlock(accessor, otherInstanceAccessor, block_if_data, typenavigator);
    }


    CppCtorPtr IntEnumClassSerializer::toCopyCtor(const std::string& name) const
    {
        CppCtorPtr c = std::make_shared<CppCtor>(name + "(const " + getFullCppTypename() + "& i)");
        std::vector<std::pair<std::string, std::string>> initList = {{"value", "i.value"}};
        c->addInitListEntries(initList);
        return c;
    }

    CppCtorPtr IntEnumClassSerializer::toInnerEnumCtor(const std::string& name) const
    {
        CppCtorPtr c = std::make_shared<CppCtor>(name + "(const " + std::string(IMPL_ENUM) + " e)");
        std::vector<std::pair<std::string, std::string>> initList = {{"value", "e"}};
        c->addInitListEntries(initList);
        return c;
    }

    CppEnumPtr IntEnumClassSerializer::toInnerEnumDefinition() const
    {
        CppEnumPtr e = std::make_shared<CppEnum>(std::string(IMPL_ENUM), "The internal enum definition of the enum of this autogenerated class.");
        for (const auto& [key, value] : typenavigator->getAcceptedValueMap())
        {
            e->addField(std::make_shared<CppEnumField>(key));
        }
        return e;
    }

    CppMethodPtr IntEnumClassSerializer::toIntMethod() const
    {
        std::stringstream doc;
        doc << "@brief int() - Converts the internally stored value to int representation \n";
        doc << "@return - the int representation";

        CppMethodPtr m = CppMethodPtr(new CppMethod("operator int() const", doc.str()));
        CppBlockPtr b = std::make_shared<CppBlock>();
        b->addLine("return EnumToValueMap.at(value);");
        m->setBlock(b);
        return m;
    }

    CppMethodPtr IntEnumClassSerializer::toCopyAssignmentMethod() const
    {
        std::stringstream doc;
        doc << "@brief operator=() -  Assignment operator for copy \n";
        doc << "@return - nothing";

        CppMethodPtr m = CppMethodPtr(new CppMethod(getFullCppTypename() + "& operator=(const " + getFullCppTypename() + "& c)", doc.str()));
        CppBlockPtr b = std::make_shared<CppBlock>();
        b->addLine("value = c.value;");
        b->addLine("return *this;");
        m->setBlock(b);
        return m;
    }

    CppMethodPtr IntEnumClassSerializer::toEnumAssignmentMethod() const
    {
        std::stringstream doc;
        doc << "@brief operator=() -  Assignment operator for the internally defined enum \n";
        doc << "@return - nothing";

        CppMethodPtr m = CppMethodPtr(new CppMethod(getFullCppTypename() + "& operator=(" + std::string(IMPL_ENUM) + " v)", doc.str()));
        CppBlockPtr b = std::make_shared<CppBlock>();
        b->addLine("value = v;");
        b->addLine("return *this;");
        m->setBlock(b);
        return m;
    }

    CppMethodPtr IntEnumClassSerializer::toIntAssignmentMethod() const
    {
        std::stringstream doc;
        doc << "@brief operator=() -  Assignment operator for the internally defined enum \n";
        doc << "@return - nothing";

        CppMethodPtr m = CppMethodPtr(new CppMethod(getFullCppTypename() + "& operator=(int v)", doc.str()));
        CppBlockPtr b = std::make_shared<CppBlock>();
        b->addLine("if (auto it = ValueToEnumMap.find(v); it == ValueToEnumMap.end())");
        CppBlockPtr b2 = std::make_shared<CppBlock>();
        b2->addLine("throw armarx::LocalException(\"The input int is not valid. Could net set the enum to value '\" + std::to_string(v) + \"'\");");
        b->addBlock(b2);
        b->addLine("else");
        CppBlockPtr b3 = std::make_shared<CppBlock>();
        b3->addLine("value = it->second;");
        b->addBlock(b3);
        b->addLine("return *this;");
        m->setBlock(b);
        return m;
    }

    CppMethodPtr IntEnumClassSerializer::toToStringMethod() const
    {
        std::stringstream doc;
        doc << "@brief toString() - Converts the internally stored value to string \n";
        doc << "@return - the name of the enum";

        CppMethodPtr m = CppMethodPtr(new CppMethod("std::string toString() const", doc.str()));
        CppBlockPtr b = std::make_shared<CppBlock>();
        b->addLine("return EnumToStringMap.at(value);");
        m->setBlock(b);
        return m;
    }

    CppMethodPtr IntEnumClassSerializer::toFromStringMethod() const
    {
        std::stringstream doc;
        doc << "@brief fromString() - sets the internally stored value to the corrsponding enum of the input str \n";
        doc << "@return - nothing";

        CppMethodPtr m = CppMethodPtr(new CppMethod("void fromString(const std::string& str)", doc.str()));
        CppBlockPtr b = std::make_shared<CppBlock>();
        b->addLine("if (auto it = StringToEnumMap.find(str); it == StringToEnumMap.end())");
        CppBlockPtr b2 = std::make_shared<CppBlock>();
        b2->addLine("throw armarx::LocalException(\"The input name is not valid. Could net set the enum to value '\" + str + \"'\");");
        b->addBlock(b2);
        b->addLine("else");
        CppBlockPtr b3 = std::make_shared<CppBlock>();
        b3->addLine("value = it->second;");
        b->addBlock(b3);

        m->setBlock(b);
        return m;
    }
}
