/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2012-2016, High Performance Humanoid Technologies (H2T),
 * Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Fabian Peller-Konrad (fabian dot peller-konrad at kit dot edu)
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

// Header
#include "ObjectClass.h"

#include <SimoxUtility/meta/type_name.h>


namespace armarx::aron::cppserializer::serializer
{
    // constructors
    ObjectClassSerializer::ObjectClassSerializer(const typenavigator::ObjectNavigatorPtr& e) :
        detail::SerializerBase<typenavigator::ObjectNavigator, ObjectClassSerializer>(
            e->getObjectName(),
            simox::meta::get_type_name<data::AronDict>(),
            simox::meta::get_type_name<type::AronObject>(),
            e)
    {
        ARMARX_CHECK_NOT_NULL(typenavigator);
        if (typenavigator->getMaybe() != type::Maybe::eNone)
        {
            throw error::MaybeNotValidException("ObjectClassSerializer", "ObjectClassSerializer", "Somehow the maybe flag of a top level object declaration is set. This is not valid!", typenavigator->getMaybe(), typenavigator->getPath());
        }
    }

    std::vector<CppFieldPtr> ObjectClassSerializer::getPublicVariableDeclarations(const std::string&) const
    {
        std::vector<CppFieldPtr> fields;
        for (const auto& [key, member] : typenavigator->getMemberTypes())
        {
            auto member_s = FromAronTypeNaviagtorPtr(member);
            std::vector<CppFieldPtr> member_fields = member_s->getPublicVariableDeclarations(key);
            fields.insert(fields.end(), member_fields.begin(), member_fields.end());
        }
        return fields;
    }

    std::vector<std::pair<std::string, std::string>> ObjectClassSerializer::getCtorInitializers(const std::string&) const
    {
        return {};
    }

    CppBlockPtr ObjectClassSerializer::getCtorBlock(const std::string&) const
    {
        CppBlockPtr b = CppBlockPtr(new CppBlock());
        return b;
    }

    CppBlockPtr ObjectClassSerializer::getResetSoftBlock(const std::string& accessor) const
    {
        CppBlockPtr block_if_data = CppBlockPtr(new CppBlock());
        if (typenavigator->getExtends() != nullptr)
        {
            const auto extends_s = FromAronTypeNaviagtorPtr(typenavigator->getExtends());
            block_if_data->addLine(extends_s->getFullCppTypename() + "::resetSoft();");
        }

        for (const auto& [key, child] : typenavigator->getMemberTypes())
        {
            auto child_s = FromAronTypeNaviagtorPtr(child);
            CppBlockPtr b2 = child_s->getResetSoftBlock(key);
            block_if_data->appendBlock(b2);
        }
        return ResolveMaybeResetSoftBlock(accessor, block_if_data, typenavigator);
    }

    CppBlockPtr ObjectClassSerializer::getResetHardBlock(const std::string& accessor) const
    {
        CppBlockPtr block_if_data = CppBlockPtr(new CppBlock());
        if (typenavigator->getExtends() != nullptr)
        {
            const auto extends_s = FromAronTypeNaviagtorPtr(typenavigator->getExtends());
            block_if_data->addLine(extends_s->getFullCppTypename() + "::resetHard();");
        }

        for (const auto& [key, child] : typenavigator->getMemberTypes())
        {
            const auto child_s = FromAronTypeNaviagtorPtr(child);
            CppBlockPtr b2 = child_s->getResetHardBlock(key);
            block_if_data->appendBlock(b2);
        }
        return ResolveMaybeResetHardBlock(accessor, block_if_data, typenavigator);
    }

    CppBlockPtr ObjectClassSerializer::getWriteTypeBlock(const std::string&) const
    {
        CppBlockPtr b = CppBlockPtr(new CppBlock());
        b->addLine("w.writeStartObject({\"" + typenavigator->getObjectName() + "\", __aronMaybeType}); // of top level object " + getCoreCppTypename());

        if (typenavigator->getExtends() != nullptr)
        {
            const auto extends_s = FromAronTypeNaviagtorPtr(typenavigator->getExtends());
            b->addLine(extends_s->getFullCppTypename() + "::writeType(w, armarx::aron::type::Maybe::eNone, true);");
            b->addLine("w.writeExtends();");
        }

        for (const auto& [key, child] : typenavigator->getMemberTypes())
        {
            const auto child_s = FromAronTypeNaviagtorPtr(child);
            b->addLine("w.writeKey(\"" + key + "\");");
            CppBlockPtr b2 = child_s->getWriteTypeBlock(child_s->getCoreCppTypename());
            b->appendBlock(b2);
        }

        b->addLine("if (!__aronExtends)");
        b->addLineAsBlock("w.writeEndObject(); // of top level object " + getCoreCppTypename());

        return b;
    }

    CppBlockPtr ObjectClassSerializer::getWriteBlock(const std::string& accessor) const
    {
        CppBlockPtr block_if_data = CppBlockPtr(new CppBlock());

        block_if_data->addLine("if (!__aronExtends)");
        block_if_data->addLineAsBlock("w.writeStartDict(); // of top level object " + getCoreCppTypename());

        if (typenavigator->getExtends() != nullptr)
        {
            const auto extends_s = FromAronTypeNaviagtorPtr(typenavigator->getExtends());
            block_if_data->addLine(extends_s->getFullCppTypename() + "::write(w, true);");
        }

        for (const auto& [key, child] : typenavigator->getMemberTypes())
        {
            const auto child_s = FromAronTypeNaviagtorPtr(child);
            CppBlockPtr child_b = CppBlockPtr(new CppBlock());

            child_b->addLine("w.writeKey(\"" + key + "\");");
            child_b->appendBlock(child_s->getWriteBlock(key));
            block_if_data->addBlock(child_b);
            block_if_data->addLine("");
        }

        block_if_data->addLine("if (!__aronExtends)");
        block_if_data->addLineAsBlock("w.writeEndDict(); // of top level object " + getCoreCppTypename());
        return ResolveMaybeWriteBlock(accessor, block_if_data, typenavigator);
    }

    CppBlockPtr ObjectClassSerializer::getReadBlock(const std::string& accessor) const
    {
        CppBlockPtr block_if_data = CppBlockPtr(new CppBlock());
        if (typenavigator->getExtends() != nullptr)
        {
            const auto extends_s = FromAronTypeNaviagtorPtr(typenavigator->getExtends());
            block_if_data->addLine(extends_s->getFullCppTypename() + "::read(r);");
        }

        block_if_data->addLine("if(!__aronIsMember)");
        block_if_data->addLineAsBlock("r.readStartDict(); // of top level object " + getCoreCppTypename());

        for (const auto& [key, child] : typenavigator->getMemberTypes())
        {
            const auto child_s = FromAronTypeNaviagtorPtr(child);
            block_if_data->addLine("r.loadMember(\"" + key + "\");");
            block_if_data->addBlock(child_s->getReadBlock(key));
        }
        block_if_data->addLine("r.readEndDict(); // of top level object " + getCoreCppTypename());
        return ResolveMaybeReadBlock(accessor, "", block_if_data, typenavigator);
    }

    CppBlockPtr ObjectClassSerializer::getEqualsBlock(const std::string& accessor, const std::string& otherInstanceAccessor) const
    {
        CppBlockPtr block_if_data = CppBlockPtr(new CppBlock());
        if (typenavigator->getExtends() != nullptr)
        {
            const auto extends_s = FromAronTypeNaviagtorPtr(typenavigator->getExtends());
            block_if_data->addLine("if (not (" + extends_s->getFullCppTypename() + "::operator== (" + otherInstanceAccessor + ")))");
            block_if_data->addLineAsBlock("return false;");
        }
        for (const auto& [key, child] : typenavigator->getMemberTypes())
        {
            auto child_s = FromAronTypeNaviagtorPtr(child);
            CppBlockPtr b2 = child_s->getEqualsBlock(key, otherInstanceAccessor + "." + key);
            block_if_data->appendBlock(b2);
        }
        return ResolveMaybeEqualsBlock(accessor, otherInstanceAccessor, block_if_data, typenavigator);
    }
}

