/*
* This file is part of ArmarX.
*
* ArmarX is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License version 2 as
* published by the Free Software Foundation.
*
* ArmarX is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program. If not, see <http://www.gnu.org/licenses/>.
*
* @author     Fabian Peller (fabian dot peller at kit dot edu)
* @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
*             GNU General Public License
*/

// STD/STL
#include <memory>
#include <numeric>

// Header
#include "NavigatorReader.h"

// ArmarX
#include <RobotAPI/libraries/aron/core/Exception.h>
#include <RobotAPI/libraries/aron/core/navigator/type/AllNavigators.h>


namespace armarx::aron::typeIO::reader
{
    NavigatorReader::NavigatorReader(const typenavigator::NavigatorPtr& n) :
        input(n)
    {
    }

    NavigatorReader::NavigatorReader(const type::AronTypePtr& n) :
        input(typenavigator::Navigator::FromAronType(n))
    {
    }

    typenavigator::NavigatorPtr NavigatorReader::getNextAndIncrease()
    {
        if (!readInitialStart)
        {
            readInitialStart = true;
            return input;
        }
        NavigatorReaderTokenPtr lastToken = stack.top();
        return lastToken->getNextElementAndIncreaseCnt();
    }

    ReaderInterface::ReadStartObjectReturnType NavigatorReader::readStartObject()
    {
        auto current_nav = getNextAndIncrease();
        auto current_nav_casted = typenavigator::ObjectNavigator::DynamicCastAndCheck(current_nav);
        auto newToken = std::make_shared<NavigatorReaderToken>(type::Descriptor::eObject, current_nav_casted);
        stack.push(newToken);
        return {newToken->getElementName(), newToken->getElementChildrenSize(), current_nav_casted->getMaybe()};
    }

    bool NavigatorReader::readEndObject()
    {
        auto token = stack.top();
        token->assertType(type::Descriptor::eObject);

        if (token->finishedElement())
        {
            stack.pop();
            return true;
        }
        return false;
    }

    ReaderInterface::ReadStartDictReturnType NavigatorReader::readStartDict()
    {
        auto current_nav = getNextAndIncrease();
        auto current_nav_casted = typenavigator::DictNavigator::DynamicCastAndCheck(current_nav);
        auto newToken = std::make_shared<NavigatorReaderToken>(type::Descriptor::eDict, current_nav_casted);
        stack.push(newToken);
        return {newToken->getElementChildrenSize(), current_nav_casted->getMaybe()};
    }

    bool NavigatorReader::readEndDict()
    {
        auto token = stack.top();
        token->assertType(type::Descriptor::eDict);

        if (token->finishedElement())
        {
            stack.pop();
            return true;
        }
        return false;
    }

    ReaderInterface::ReadStartListReturnType NavigatorReader::readStartList()
    {
        auto current_nav = getNextAndIncrease();
        auto current_nav_casted = typenavigator::ListNavigator::DynamicCastAndCheck(current_nav);
        auto newToken = std::make_shared<NavigatorReaderToken>(type::Descriptor::eList, current_nav_casted);
        stack.push(newToken);
        return {newToken->getElementChildrenSize(), current_nav_casted->getMaybe()};
    }

    bool NavigatorReader::readEndList()
    {
        auto token = stack.top();
        token->assertType(type::Descriptor::eList);

        if (token->finishedElement())
        {
            stack.pop();
            return true;
        }
        return false;
    }

    ReaderInterface::ReadStartTupleReturnType NavigatorReader::readStartTuple()
    {
        auto current_nav = getNextAndIncrease();
        auto current_nav_casted = typenavigator::TupleNavigator::DynamicCastAndCheck(current_nav);
        auto newToken = std::make_shared<NavigatorReaderToken>(type::Descriptor::eTuple, current_nav_casted);
        stack.push(newToken);
        return {newToken->getElementChildrenSize(), current_nav_casted->getMaybe()};
    }

    bool NavigatorReader::readEndTuple()
    {
        auto token = stack.top();
        token->assertType(type::Descriptor::eTuple);

        if (token->finishedElement())
        {
            stack.pop();
            return true;
        }
        return false;
    }

    ReaderInterface::ReadStartPairReturnType NavigatorReader::readStartPair()
    {
        auto current_nav = getNextAndIncrease();
        auto current_nav_casted = typenavigator::PairNavigator::DynamicCastAndCheck(current_nav);
        auto newToken = std::make_shared<NavigatorReaderToken>(type::Descriptor::ePair, current_nav_casted);
        stack.push(newToken);
        return {newToken->getElementChildrenSize(), current_nav_casted->getMaybe()};
    }

    bool NavigatorReader::readEndPair()
    {
        auto token = stack.top();
        token->assertType(type::Descriptor::ePair);

        if (token->finishedElement())
        {
            stack.pop();
            return true;
        }
        return false;
    }

    ReaderInterface::ReadEigenMatrixReturnType NavigatorReader::readEigenMatrix()
    {
        auto nav = getNextAndIncrease();
        auto casted = typenavigator::EigenMatrixNavigator::DynamicCastAndCheck(nav);
        return {casted->getRows(), casted->getCols(), casted->getTypename(), casted->getMaybe()};
    }

    ReaderInterface::ReadEigenQuaternionReturnType NavigatorReader::readEigenQuaternion()
    {
        auto nav = getNextAndIncrease();
        auto casted = typenavigator::EigenMatrixNavigator::DynamicCastAndCheck(nav);
        return {casted->getTypename(), casted->getMaybe()};
    }

    ReaderInterface::ReadIVTCbyteImageReturnType NavigatorReader::readIVTCByteImage()
    {
        auto nav = getNextAndIncrease();
        auto casted = typenavigator::IVTCByteImageNavigator::DynamicCastAndCheck(nav);
        return {casted->getMaybe()};
    }

    ReaderInterface::ReadOpenCVMatReturnType NavigatorReader::readOpenCVMat()
    {
        auto nav = getNextAndIncrease();
        auto casted = typenavigator::OpenCVMatNavigator::DynamicCastAndCheck(nav);
        return {casted->getMaybe()};
    }

    ReaderInterface::ReadPCLPointCloudReturnType NavigatorReader::readPCLPointCloud()
    {
        auto nav = getNextAndIncrease();
        auto casted = typenavigator::PCLPointCloudNavigator::DynamicCastAndCheck(nav);
        return {casted->getTypename(), casted->getMaybe()};
    }

    ReaderInterface::ReadPositionReturnType NavigatorReader::readPosition()
    {
        auto nav = getNextAndIncrease();
        auto casted = typenavigator::PositionNavigator::DynamicCastAndCheck(nav);
        return {casted->getMaybe()};
    }

    ReaderInterface::ReadOrientationReturnType NavigatorReader::readOrientation()
    {
        auto nav = getNextAndIncrease();
        auto casted = typenavigator::OrientationNavigator::DynamicCastAndCheck(nav);
        return {casted->getMaybe()};
    }

    ReaderInterface::ReadPoseReturnType NavigatorReader::readPose()
    {
        auto nav = getNextAndIncrease();
        auto casted = typenavigator::PoseNavigator::DynamicCastAndCheck(nav);
        return {casted->getMaybe()};
    }

    // Read primitives
    ReaderInterface::ReadPrimitiveReturnType NavigatorReader::readInt()
    {
        auto nav = getNextAndIncrease();
        auto casted = typenavigator::IntNavigator::DynamicCastAndCheck(nav);
        return {casted->getMaybe()};
    }

    ReaderInterface::ReadPrimitiveReturnType NavigatorReader::readFloat()
    {
        auto nav = getNextAndIncrease();
        auto casted = typenavigator::FloatNavigator::DynamicCastAndCheck(nav);
        return {casted->getMaybe()};
    }

    ReaderInterface::ReadPrimitiveReturnType NavigatorReader::readLong()
    {
        auto nav = getNextAndIncrease();
        auto casted = typenavigator::LongNavigator::DynamicCastAndCheck(nav);
        return {casted->getMaybe()};
    }

    ReaderInterface::ReadPrimitiveReturnType NavigatorReader::readDouble()
    {
        auto nav = getNextAndIncrease();
        auto casted = typenavigator::DoubleNavigator::DynamicCastAndCheck(nav);
        return {casted->getMaybe()};
    }

    ReaderInterface::ReadPrimitiveReturnType NavigatorReader::readString()
    {
        auto nav = getNextAndIncrease();
        auto casted = typenavigator::StringNavigator::DynamicCastAndCheck(nav);
        return {casted->getMaybe()};
    }

    ReaderInterface::ReadPrimitiveReturnType NavigatorReader::readBool()
    {
        auto nav = getNextAndIncrease();
        auto casted = typenavigator::BoolNavigator::DynamicCastAndCheck(nav);
        return {casted->getMaybe()};
    }

    ReaderInterface::ReadPrimitiveReturnType NavigatorReader::readTime()
    {
        auto nav = getNextAndIncrease();
        auto casted = typenavigator::TimeNavigator::DynamicCastAndCheck(nav);
        return {casted->getMaybe()};
    }

    std::string NavigatorReader::readKey()
    {
        auto token = stack.top();
        return token->getCurrentKey();
    }

    void NavigatorReader::loadMember(const std::string& k)
    {
        auto token = stack.top();
        token->setCurrentKey(k);
    }

    // Helper functions
    type::Descriptor NavigatorReader::getTypeOfNext() const
    {
        auto token = stack.top();
        return token->getTypeOfNext();
    }
}
