/*
* This file is part of ArmarX.
*
* ArmarX is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License version 2 as
* published by the Free Software Foundation.
*
* ArmarX is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program. If not, see <http://www.gnu.org/licenses/>.
*
* @author     Fabian Peller (fabian dot peller at kit dot edu)
* @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
*             GNU General Public License
*/

#include "NavigatorWriter.h"

#include <RobotAPI/libraries/aron/core/navigator/data/AllNavigators.h>
#include <RobotAPI/libraries/aron/core/navigator/type/AllNavigators.h>

#include <memory>
#include <numeric>


namespace armarx::aron::typeIO::writer
{
    // generate Path info
    Path NavigatorWriter::generatePath()
    {
        Path path;
        if (!wroteInitialStartObject)
        {
            wroteInitialStartObject = true;
        }
        else
        {
            auto current = stack.top();
            path = Path(current->getElement()->getPath(), current->toAccessor());
        }
        return path;
    }

    // interface
    void NavigatorWriter::writeStartObject(const WriteStartObjectInput& o)
    {
        Path path = generatePath();
        auto type = std::make_shared<typenavigator::ObjectNavigator>(path);
        type->setObjectName(o.name);
        type->setMaybe(o.maybe);
        auto new_token = std::make_shared<NavigatorWriterToken>(type->getDescriptor(), type);
        stack.push(new_token);
    }

    void NavigatorWriter::writeEndObject()
    {
        lastRemovedToken = stack.top();
        stack.pop();

        if (stack.size() > 0)
        {
            auto prevToken = stack.top();
            prevToken->addElement(lastRemovedToken->getElement());
        }
    }

    void NavigatorWriter::writeStartDict(const WriteStartDictInput& o)
    {
        Path path = generatePath();
        auto type = std::make_shared<typenavigator::DictNavigator>(path);
        type->setMaybe(o.maybe);
        auto new_token = std::make_shared<NavigatorWriterToken>(type->getDescriptor(), type);
        stack.push(new_token);
    }

    void NavigatorWriter::writeEndDict()
    {
        lastRemovedToken = stack.top();
        stack.pop();

        if (stack.size() > 0)
        {
            auto prevToken = stack.top();
            prevToken->addElement(lastRemovedToken->getElement());
        }
    }

    void NavigatorWriter::writeStartList(const WriteStartListInput& o)
    {
        Path path = generatePath();
        auto type = std::make_shared<typenavigator::ListNavigator>(path);
        type->setMaybe(o.maybe);
        auto new_token = std::make_shared<NavigatorWriterToken>(type->getDescriptor(), type);
        stack.push(new_token);
    }

    void NavigatorWriter::writeEndList()
    {
        lastRemovedToken = stack.top();
        stack.pop();

        if (stack.size() > 0)
        {
            auto prevToken = stack.top();
            prevToken->addElement(lastRemovedToken->getElement());
        }
    }

    void NavigatorWriter::writeStartTuple(const WriteStartTupleInput& o)
    {
        Path path = generatePath();
        auto type = std::make_shared<typenavigator::TupleNavigator>(path);
        type->setMaybe(o.maybe);
        auto new_token = std::make_shared<NavigatorWriterToken>(type->getDescriptor(), type);
        stack.push(new_token);
    }

    void NavigatorWriter::writeEndTuple()
    {
        lastRemovedToken = stack.top();
        stack.pop();

        if (stack.size() > 0)
        {
            auto prevToken = stack.top();
            prevToken->addElement(lastRemovedToken->getElement());
        }
    }

    void NavigatorWriter::writeStartPair(const WriteStartPairInput& o)
    {
        Path path = generatePath();
        auto type = std::make_shared<typenavigator::PairNavigator>(path);
        type->setMaybe(o.maybe);
        auto new_token = std::make_shared<NavigatorWriterToken>(type->getDescriptor(), type);
        stack.push(new_token);
    }

    void NavigatorWriter::writeEndPair()
    {
        lastRemovedToken = stack.top();
        stack.pop();

        if (stack.size() > 0)
        {
            auto prevToken = stack.top();
            prevToken->addElement(lastRemovedToken->getElement());
        }
    }


    template <class ElementT, class InputT>
    std::shared_ptr<ElementT>
    NavigatorWriter::writeElement(const InputT& input)
    {
        Path path = generatePath();
        NavigatorWriterTokenPtr token = stack.top();
        auto type = std::make_shared<ElementT>(path);
        type->setMaybe(input.maybe);
        token->addElement(type);

        return type;
    }


    void NavigatorWriter::writeEigenMatrix(const WriteEigenMatrixInput& o)
    {
        auto type = writeElement<typenavigator::EigenMatrixNavigator>(o);

        type->setRows(o.rows);
        type->setCols(o.cols);
        type->setTypename(o.type);
    }

    void NavigatorWriter::writeEigenQuaternion(const WriteEigenQuaternionInput& o)
    {
        auto type = writeElement<typenavigator::EigenQuaternionNavigator>(o);

        type->setTypename(o.type);
    }


    void NavigatorWriter::writeImage(const WriteImageInput& o)
    {
        auto type = writeElement<typenavigator::ImageNavigator>(o);

        type->setPixelType(o.pixelType);
    }


    void NavigatorWriter::writeIVTCByteImage(const WriteIVTCByteImageInput& o)
    {
        writeElement<typenavigator::IVTCByteImageNavigator>(o);
    }

    void NavigatorWriter::writeOpenCVMat(const WriteOpenCVMatInput& o)
    {
        writeElement<typenavigator::OpenCVMatNavigator>(o);
    }

    void NavigatorWriter::writePCLPointCloud(const WritePCLPointCloudInput& o)
    {
        auto type = writeElement<typenavigator::PCLPointCloudNavigator>(o);

        type->setTypename(o.type);
    }

    void NavigatorWriter::writePosition(const WritePositionInput& o)
    {
        writeElement<typenavigator::PositionNavigator>(o);
    }

    void NavigatorWriter::writeOrientation(const WriteOrientationInput& o)
    {
        writeElement<typenavigator::OrientationNavigator>(o);
    }

    void NavigatorWriter::writePose(const WritePoseInput& o)
    {
        writeElement<typenavigator::PoseNavigator>(o);
    }

    void NavigatorWriter::writeInt(const WritePrimitiveInput& o)
    {
        writeElement<typenavigator::IntNavigator>(o);
    }

    void NavigatorWriter::writeLong(const WritePrimitiveInput& o)
    {
        writeElement<typenavigator::LongNavigator>(o);
    }

    void NavigatorWriter::writeFloat(const WritePrimitiveInput& o)
    {
        writeElement<typenavigator::FloatNavigator>(o);
    }

    void NavigatorWriter::writeDouble(const WritePrimitiveInput& o)
    {
        writeElement<typenavigator::DoubleNavigator>(o);
    }

    void NavigatorWriter::writeString(const WritePrimitiveInput& o)
    {
        writeElement<typenavigator::StringNavigator>(o);
    }

    void NavigatorWriter::writeBool(const WritePrimitiveInput& o)
    {
        writeElement<typenavigator::BoolNavigator>(o);
    }

    void NavigatorWriter::writeTime(const WritePrimitiveInput& o)
    {
        writeElement<typenavigator::TimeNavigator>(o);
    }

    void NavigatorWriter::writeKey(const std::string& k)
    {
        auto token = stack.top();
        token->setCurrentKey(k);
    }

    void NavigatorWriter::writeExtends()
    {
        if (stack.size() < 2)
        {
            throw error::SizeNotValidException("NavigatorWriter", "writeExtends", "The stacksize must at least be 2 (one for base, one for derived)", stack.size(), 2);
        }

        auto base = stack.top();
        stack.pop();
        auto derived = stack.top();

        auto base_obj = typenavigator::ObjectNavigator::DynamicCastAndCheck(base->getElement());
        auto derived_obj = typenavigator::ObjectNavigator::DynamicCastAndCheck(derived->getElement());
        derived_obj->setExtends(base_obj);
    }
}
